@@ -30,8 +30,7 @@ Net::FTP - FTP Client class
 
 =head1 DESCRIPTION
 
-C<Net::FTP> is a class implementing a simple FTP client in Perl as described
-in RFC959
+C<Net::FTP> is a class implementing a simple FTP client in Perl.
 
 =head2 TO BE CONTINUED ...
 
@@ -40,11 +39,11 @@ in RFC959
 require 5.001;
 use Socket 1.3;
 use Carp;
-use Net::Socket;
+use IO::Socket;
 
-@ISA = qw(Net::Socket);
+@ISA = qw(IO::Socket);
 
-$VERSION = sprintf("%d.%02d", q$Revision: 1.18 $ =~ /(\d+)\.(\d+)/);
+$VERSION = sprintf("%d.%02d", q$Revision: 1.8 $ =~ /(\d+)\.(\d+)/);
 sub Version { $VERSION }
 
 use strict;
@@ -64,20 +63,19 @@ remote host. Possible options are:
 
 =cut
 
-sub FTP_READY    { 0 } # Ready 
-sub FTP_RESPONSE { 1 } # Waiting for a response
-sub FTP_XFER     { 2 } # Doing data xfer
+sub FTP_READY    { 0 }
+sub FTP_RESPONSE { 1 }
+sub FTP_XFER     { 2 }
 
 sub new {
  my $pkg  = shift;
  my $host = shift;
  my %arg  = @_; 
- my $me = bless Net::Socket->new(Peer	=> $host, 
+ my $me = bless IO::Socket->new(Peer	=> $host, 
 				Service	=> 'ftp', 
 				Port	=> $arg{Port} || 'ftp'
 				), $pkg;
 
- ${*$me} = "";					# partial response text
  @{*$me} = ();					# Last response text
 
  %{*$me} = (%{*$me},				# Copy current values
@@ -97,9 +95,6 @@ sub new {
 
  $me->autoflush(1);
 
- $me->debug($arg{Debug})
-   if(exists $arg{Debug});
-
  unless(2 == $me->response())
   {
    $me->close();
@@ -125,13 +120,8 @@ sub debug {
  my $me = shift;
  my $debug = ${*$me}{Debug};
  
- if(@_)
-  {
-   ${*$me}{Debug} = 0 + shift;
-
-   printf STDERR "\n$me VERSION %s\n", $Net::FTP::VERSION
-     if(${*$me}{Debug});
-  }
+ ${*$me}{Debug} = 0 + shift
+	if @_;
 
  $debug;
 }
@@ -193,14 +183,8 @@ sub login {
  my $acct = shift if(defined $pass);
  my $ok;
 
- unless(defined $user)
-  {
-   require Net::Netrc;
-   my $rc = Net::Netrc->lookup(${*$me}{FtpHost});
-
-   ($user,$pass,$acct) = $rc->lpa()
-	if $rc;
-  }
+ ($user,$pass,$acct) = netrc(${*$me}{FtpHost})
+	unless defined $user;
 
  $user = "anonymous"
 	unless defined $user;
@@ -281,7 +265,6 @@ sub get {
  my $me = shift;
  my $remote = shift;
  my $local  = shift;
- my $where  = shift || 0;
  my($loc,$len,$buf,$resp,$localfd,$data);
  local *FD;
 
@@ -298,21 +281,15 @@ sub get {
   {
    $loc = \*FD;
 
-   unless(($where) ? open($loc,">>$local") : open($loc,">$local"))
+   unless(open($loc,">$local"))
     {
      carp "Cannot open Local file $local: $!\n";
      return undef;
     }
   }
 
- if ($where) {   
-   $data = $me->rest_cmd($where,$remote) or
-	return undef; 
- }
- else {
-   $data = $me->retr($remote) or
-     return undef;
- }
+ $data = $me->retr($remote) or
+	return undef;
 
  $buf = '';
 
@@ -325,7 +302,9 @@ sub get {
  close($loc)
 	unless $localfd;
  
- $data->close() == 2; # implied $me->response
+ $resp = $data->close();
+
+ 200 <= $resp && $resp < 300;
 }
 
 sub cwd {
@@ -336,13 +315,6 @@ sub cwd {
 		     : $me->CWD($dir);
 }
 
-sub pwd {
- my $me = shift;
-
- $me->PWD() ? ($me->message =~ /\"([^\"]+)/)[0]
-            : undef;
-}
-
 sub put	       { shift->send("stor",@_) }
 sub put_unique { shift->send("stou",@_) }
 sub append     { shift->send("appe",@_) }
@@ -427,7 +399,7 @@ sub port {
 
    # create a Listen socket at same address as the command socket
 
-   $listen = Net::Socket->new(Listen  => 5,
+   $listen = IO::Socket->new(Listen  => 5,
 			     Service => 'ftp',
 			     Addr    => $me->sockhost, 
 			    );
@@ -476,7 +448,7 @@ sub timeout {
 
 sub accept {
  my $me = shift;
-
+ 
  return undef unless defined ${*$me}{LISTEN};
 
  my $data = ${*$me}{LISTEN}->accept;
@@ -515,30 +487,25 @@ sub list_cmd {
  my $cmd = lc shift;
  my $data = $me->$cmd(@_);
 
- return undef
-	unless(defined $data);
+ die "undef" unless(defined $data);
 
  bless $data, "Net::FTP::A"; # Force ASCII mode
 
  my $databuf = '';
- my $buf = '';
+ my $list = [];
 
  while($data->read($databuf,1024)) {
-   $buf .= $databuf;
+   push(@{$list}, split(/\n/,$databuf)); ## break into lines
  }
 
- my $list = [ split(/\n/,$buf) ];
-
- $data->close();
-
  wantarray ? @{$list} : $list;
 }
 
 sub data_cmd {
  my $me = shift;
- my $cmd = uc shift;
  my $ok = 1;
  my $pasv = defined ${*$me}{Pasv} ? 1 : 0;
+ my $cmd = uc shift;
 
  $ok = $me->port
 	unless $pasv || defined ${*$me}{Port};
@@ -551,27 +518,6 @@ sub data_cmd {
 		    : undef;
 }
 
-sub rest_cmd {
- my $me = shift;
- my $ok = 1;
- my $pasv = defined ${*$me}{Pasv} ? 1 : 0;
- my $where = shift;
- my $file = shift;
-
- $ok = $me->port
-	unless $pasv || defined ${*$me}{Port};
-
- $ok = $me->REST($where)
-	if $ok;
-
- $ok = $me->RETR($file)
-	if $ok;
-
- return $pasv ? $ok
-	      : $ok ? $me->accept()
-		    : undef;
-}
-
 sub cmd {
  my $me = shift;
 
@@ -592,7 +538,7 @@ sub send_cmd {
 
   ${*$me}{State} = FTP_RESPONSE;
 
-  printf STDERR "\n$me>> %s", $cmd=~/^(pass|resp)/i ? "$1 ....\n" : $cmd
+  printf STDERR "$me>> %s\n", $cmd=~/^(pass|resp)/i ? "$1 ...." : $cmd
 	if $me->debug;
  }
 
@@ -629,16 +575,14 @@ sub response {
  my($code,$more,$rin,$rout,$partial,$buf) = (undef,0,'','','','');
 
  @{*$me} = (); # the responce
- $buf = ${*$me};
- my @buf = ();
 
  vec($rin,fileno($me),1) = 1;
 
  do
   {
-   if(length($buf) || ($timeout==0) || select($rout=$rin, undef, undef, $timeout))
+   if(($timeout==0) || select($rout=$rin, undef, undef, $timeout))
     {
-     unless(length($buf) || sysread($me, $buf, 1024))
+     unless(sysread($me, $buf, 1024))
       {
        carp "Unexpected EOF on command channel";
        return undef;
@@ -646,25 +590,21 @@ sub response {
 
      substr($buf,0,0) = $partial;    ## prepend from last sysread
 
-     @buf = split(/\r?\n/, $buf);  ## break into lines
+     my @buf = split(/\r?\n/, $buf);  ## break into lines
 
      $partial = (substr($buf, -1, 1) eq "\n") ? ''
 					      : pop(@buf); 
 
-     $buf = "";
-
-     while (@buf)
+     my $cmd;
+     foreach $cmd (@buf)
       {
-       my $cmd = shift @buf;
        print STDERR "$me<< $cmd\n"
 	 if $me->debug;
  
        ($code,$more) = ($1,$2)
 	if $cmd =~ /^(\d\d\d)(.)/;
-
+ 
        push(@{*$me},$');
-
-       last unless(defined $more && $more eq "-");
       } 
     }
    else
@@ -673,10 +613,7 @@ sub response {
      return undef;
     }
   }
- while((scalar(@{*$me}) == 0) || (defined $more && $more eq "-"));
-
- ${*$me} = @buf ? join("\n",@buf,"") : "";
- ${*$me} .= $partial;
+ while(length($partial) || (defined $more && $more eq "-"));
 
  ${*$me}{Code} = $code;
  ${*$me}{State} = FTP_READY;
@@ -684,6 +621,138 @@ sub response {
  substr($code,0,1);
 }
 
+sub netrc {
+ my $host = shift;
+ my $file = (getpwuid($>))[7] . "/.netrc";
+ my($login,$pass,$acct) = (undef,undef,undef);
+ local *NETRC;
+ local $_;
+
+ my @stat = stat($file);
+
+ if(@stat)
+  {
+   if($stat[2] & 077)
+    {
+     carp "Bad permissions: $file";
+     return ();
+    }
+   if($stat[4] != $<)
+    {
+     carp "Not owner: $file";
+     return ();
+    }
+  }
+
+ if(open(NETRC,$file))
+  {
+   my($mach,$macdef,$tok,@tok) = (0,0);
+
+LINE_LOOP:
+   while(<NETRC>) 
+    {
+     $macdef = 0 if /\A\n\Z/;
+
+     next if $macdef;
+
+     push(@tok, split(/[\s\n]+/, $_));
+
+TOKEN_LOOP:
+     while(@tok)
+      {
+       if($tok[0] eq "default")
+	{
+	 last LINE_LOOP if $mach;
+
+	 shift(@tok);
+	 $mach = 1;
+	 ($login,$pass,$acct) = (undef,undef,undef);
+
+	 next TOKEN_LOOP;
+	}
+
+       last TOKEN_LOOP unless @tok > 1;
+       $tok = shift(@tok);
+
+       if($tok eq "machine")
+	{
+	 last LINE_LOOP if $mach;
+	 $mach = 1 if $host eq shift(@tok);
+	 ($login,$pass,$acct) = (undef,undef,undef);
+	}
+       elsif($tok eq "login")
+	{
+	 $login = shift(@tok);
+	}
+       elsif($tok eq "password")
+	{
+	 $pass = shift(@tok);
+	}
+       elsif($tok eq "account")
+	{
+	 $acct = shift(@tok);
+	}
+       elsif($tok eq "macdef")
+	{
+	 $macdef = 1;
+	}
+      }
+    }
+   close(NETRC);
+
+   return ($login,$pass,$acct)
+	if $mach;
+  }
+
+ return ();
+}
+
+sub file_mode {
+ local $_ = shift;
+ my $mode = 0;
+ my($type,$ch);
+
+ s/^(.)// and $type = $1;
+
+ foreach $ch (split(//,$_))
+  {
+   $mode <<= 1;
+   $mode |= 1 unless $ch eq "-";
+  }
+
+ $type eq "d" and $mode |= 0040000 or	# Directory
+   $type eq "l" and $mode |= 0120000 or	# Symbolic Link
+   $mode |= 0100000;			# Regular File
+
+ $mode |= 0004000 if /^...s....../i;
+ $mode |= 0002000 if /^......s.../i;
+ $mode |= 0001000 if /^.........t/i;
+
+ $mode;
+}
+
+sub parse_dir
+{
+ my $me = shift;
+ my $dir = shift;
+ my @files = ();
+
+ local $_;
+
+ foreach (@$dir)
+  {
+   if(/^([\-FlrwxsStTdD]{10}).*(\w+)\s+(\w*\D)\s*(\d+)\s+(\w{3}\s+\d+\s*(\d+:\d+|\d{4}))\s+(\S+)(\s+->\s+(\S+))?/ )
+    {
+     my($mode,$owner,$group,$size,$date,$file,$link) = ($1,$2,$3,$4,$5,$7,$9);
+
+     $mode = file_mode($mode);
+     push(@files, [$mode, $owner, $group, $size, $date, $file, $link]);
+    }
+  }
+ wantarray ? @files : \@files;
+}
+
+
 ;########################################
 ;#
 ;# RFC959 commands
@@ -713,7 +782,6 @@ sub RNFR { shift->send_cmd("RNFR",@_)->response() == 3}
 sub RNTO { shift->send_cmd("RNTO",@_)->response() == 2}
 sub ACCT { shift->send_cmd("ACCT",@_)->response() == 2}
 sub RESP { shift->send_cmd("RESP",@_)->response() == 2}
-sub REST { shift->send_cmd("REST",@_)->response() == 3}
 sub USER { my $ok = shift->send_cmd("USER",@_)->response();($ok == 2 || $ok == 3) ? $ok : 0;}
 sub PASS { my $ok = shift->send_cmd("PASS",@_)->response();($ok == 2 || $ok == 3) ? $ok : 0;}
 sub AUTH { my $ok = shift->send_cmd("AUTH",@_)->response();($ok == 2 || $ok == 3) ? $ok : 0;}
@@ -727,10 +795,10 @@ sub SYST { no_imp; }
 sub STAT { no_imp; }
 sub STRU { no_imp; }
 sub REIN { no_imp; }
+sub REST { no_imp; }
 
 package Net::FTP::dataconn;
 use Carp;
-no strict 'vars';
 
 sub abort {
  my $fd = shift;
@@ -744,24 +812,15 @@ sub close {
  my $fd = shift;
  my $ftp = ${*$fd}{Cmd};
 
- $fd->Net::Socket::close();
+ $fd->IO::Socket::close();
  delete ${*$ftp}{DATA};
 
  $ftp->response();
 }
 
-sub timeout {
- my $me = shift;
- my $timeout = ${*$me}{Timeout};
-
- ${*$me}{Timeout} = 0 + shift if(@_);
-
- $timeout;
-}
-
 sub _select {
  my $fd = shift;
- local *timeout = \$_[0]; shift;
+ my $timeout = shift;
  my $rw = shift;
  my($rin,$win);
 
@@ -783,14 +842,14 @@ sub _select {
 
 sub can_read {
  my $fd = shift;
- local *timeout = \$_[0];
+ my $timeout = shift;
 
  $fd->_select($timeout,1);
 }
 
 sub can_write {
  my $fd = shift;
- local *timeout = \$_[0];
+ my $timeout = shift;
 
  $fd->_select($timeout,0);
 }
@@ -801,6 +860,8 @@ sub cmd {
  ${*$me}{Cmd};
 }
 
+# should I have a close sub which calls response ??
+
 
 @Net::FTP::L::ISA = qw(Net::FTP::I);
 @Net::FTP::E::ISA = qw(Net::FTP::I);
@@ -814,54 +875,29 @@ no strict 'vars';
 sub read {
  my $fd = shift;
  local *buf = \$_[0]; shift;
- my $size = shift || croak 'read($buf,$size,[$offset])';
- my $offset = shift || 0;
- my $timeout = ${*$fd}{Timeout};
- my $l;
+ my $size = shift || croak 'read($buf,$size,[$timeout])';
+ my $timeout = @_ ? shift : ${*$fd}{Timeout};
 
- croak "Bad offset"
-	if($offset < 0);
-
- $offset = length $buf
-	if($offset > length $buf);
-
- $l = 0;
- READ:
-  {
-   $fd->can_read($timeout) or
+ $fd->can_read($timeout) or
 	croak "Timeout";
 
-   my $n = sysread($fd, ${*$fd}, $size, length ${*$fd});
+ # this needs a bit more thought so I return the correct number of bytes !!
 
-   return $n
-	unless($n >= 0);
+ $buf = '';
 
-#   my $lf = substr(${*$fd},-1,1) eq "\r" ? chop(${*$fd})
-#					 : "";
+ my $n = sysread($fd, $buf, $size);
 
-   my $lf = (length ${*$fd} > 0 && substr(${*$fd},-1,1) eq "\r") ? chop(${*$fd})
-                     : "";
+ if($n >= 0)
+  {
+   substr($buf,0,0) = ${*$fd};
+   $buf =~ s/\r\n/\n/g;
+   $buf =~ s/([^\n]*)\Z//so;
+   ${*$fd} = $1;
 
-   ${*$fd} =~ s/\r\n/\n/go;
-
-   substr($buf,$offset) = ${*$fd};
-
-   $l += length(${*$fd});
-   $offset += length(${*$fd});
-
-   ${*$fd} = $lf;
-   
-   redo READ
-     if($l == 0 && $n > 0);
-
-   if($n == 0 && $l == 0)
-    {
-     substr($buf,$offset) = ${*$fd};
-     ${*$fd} = "";
-    }
+   $n = length $buf;
   }
 
- return $l;
+ $n;
 }
 
 sub write {
@@ -928,7 +964,7 @@ Graham Barr <Graham.Barr@tiuk.ti.com>
 
 =head2 REVISION
 
-$Revision: 1.17 $
+$Revision: 1.8 $
 
 =head2 COPYRIGHT
 
